#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "config.h"
#include "rzb_global.h"
#include "rzb_network.h"
#include "rzb_client.h"

#define BACKLOG 10

typedef struct _RZBNetworkConnection
{
    int sock;
    unsigned long bytes_in;
    unsigned long bytes_out;
} RZBNC;

/*
 *
 * Wrapper for establishing a local listener
 *
 */
SO_PUBLIC int MakeListener(const char * port)
{
    struct addrinfo hints;
    struct addrinfo *servinfo;
    int one = 1;
    struct linger linger = {1, 1};
    int sock;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE;

    if (getaddrinfo(NULL, port, &hints, &servinfo) != 0)
    {
        perror("getaddrinfo error 1");
        exit(-1);
    }

    //Create Socket
    if ((sock = socket(servinfo->ai_family, servinfo->ai_socktype, servinfo->ai_protocol)) == -1)
    {
        perror("Socket create error");
        exit(-1);
    }

    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
    setsockopt(sock, SOL_SOCKET, SO_LINGER, &linger, sizeof(linger));

    //Bind to local port
    if (bind(sock, servinfo->ai_addr, servinfo->ai_addrlen) == -1)
    {
        perror("Bind error");
        exit(-1);
    }

    //Listen for connections
    if (listen(sock, BACKLOG) == -1)
    {
        perror("Listen error");
        exit(-1);
    }

    //servinfo no longer need, so free it
    //then return socket
    freeaddrinfo(servinfo);

    return sock;
}

SO_PUBLIC RZBNC *MakeConnection(const int sock)
{
    RZBNetworkConnection *net;

    if ((net = calloc(1, sizeof(*net))) == NULL)
    {
        perror("failed to allocate a network connection");
        exit(-1);
    }
    net->sock = sock;
    return net;
}

/*
 *
 * Wrapper for establishing a connection with a remote interface
 *
 */
SO_PUBLIC RZBNC *ConnectMe(const char *ipaddr, const char *port)
{
    struct addrinfo hints;
    struct addrinfo *servinfo;
    RZBNetworkConnection *net;
    int rval;
    unsigned to_secs = 10;

    memset(&hints, 0, sizeof(struct addrinfo));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE;

    if (getaddrinfo(ipaddr, port, &hints, &servinfo) != 0)
    {
        perror("getaddrinfo error 2");
        return NULL;
    }

    if ((net = calloc(1, sizeof(*net))) == NULL)
    {
        perror("failed to allocate a network connection");
        freeaddrinfo(servinfo);
        exit(-1);
    }

    //Create Socket
    if ((net->sock = socket(servinfo->ai_family, servinfo->ai_socktype, servinfo->ai_protocol)) == -1)
    {
        free(net);
        freeaddrinfo(servinfo);
        perror("Socket create error");
        return NULL;
    }

    //Establish a connection
    do
    {
        if ((rval = connect(net->sock, servinfo->ai_addr, servinfo->ai_addrlen)))
        {
            fprintf(stderr, "Error connecting to %s:%s: (%d) %s\n", ipaddr, port, errno, strerror(errno));
            to_secs--;
            sleep(1);
        }
    } while (rval && to_secs);

    freeaddrinfo(servinfo);

    if (rval)
    {
        free(net);
        return NULL;
    }

    return net;
}


SO_PUBLIC void NetworkConnectionDestroy(RZBNC *net)
{
    if (net)
    {
        if (net->sock != -1)
        {
            close(net->sock);
        }
        free(net);
    }
}

SO_PUBLIC void NetworkConnectionGetStats(RZBNC *net, unsigned long *bytes_in, unsigned long *bytes_out)
{
    *bytes_in = net->bytes_in;
    *bytes_out = net->bytes_out;
}

/*
 *
 * Wrapper functions for TCPSocket send/recv calls
 *
 */
SO_PUBLIC HRESULT sendWrap(RZBNC *net, const void *buf, unsigned bufsize)
{
    int bytessofar;
    unsigned totalbytes;
    int sock = net->sock;
    uint8_t *buffer = (uint8_t *)buf;

    if (sock == -1)
        return R_FAIL;

    totalbytes = 0;
    while (totalbytes < bufsize)
    {
        bytessofar = send(sock, buffer+totalbytes, bufsize-totalbytes, 0);
        if (bytessofar < 1)
        {
            fprintf(stderr, "Could not send data: (%d) %s\n", errno, strerror(errno));
            return R_FAIL;
        }
        totalbytes += (unsigned)bytessofar;
    }

    net->bytes_out += (unsigned long)totalbytes;
    return R_SUCCESS;
}

SO_PUBLIC HRESULT recvWrap(RZBNC *net, void *buf, unsigned bufsize)
{
    int sock = net->sock;
    int bytessofar;
    unsigned totalbytes;
    uint8_t *buffer = (uint8_t *)buf;

    totalbytes = 0;
    while (totalbytes < bufsize)
    {
        bytessofar = recv(sock, buffer+totalbytes, bufsize-totalbytes, 0);
        if (bytessofar < 1)
        {
            fprintf(stderr, "Could not receive data: (%d) %s\n", errno, strerror(errno));
            return R_FAIL;
        }
        totalbytes += (unsigned)bytessofar;
    }

    net->bytes_in += (unsigned long)totalbytes;
    return R_SUCCESS;
}

/*
 *
 * Protocol sanity checks
 *
 */
SO_PUBLIC void protocolFill(SNAKECHARMER *scpacket)
{
    memset(scpacket, 0, sizeof(*scpacket));
    memcpy(scpacket->magic, "NRT", 3);
    scpacket->version = 1;
}

SO_PUBLIC HRESULT acknowledge(RZBNC *net, SNAKECHARMER *scpacket)
{
    protocolFill(scpacket);
    scpacket->msgtype = ACK;

    if ((sendWrap(net, scpacket, sizeof(*scpacket))) != R_SUCCESS)
        return R_FAIL;

    return R_SUCCESS;
}

SO_PUBLIC HRESULT sendFailure(RZBNC *net, SNAKECHARMER *scpacket)
{
    protocolFill(scpacket);
    scpacket->msgtype = RST;

    if ((sendWrap(net, scpacket, sizeof(*scpacket))) != R_SUCCESS)
        return R_FAIL;

    return R_SUCCESS;
}


SO_PUBLIC HRESULT protocolCheck(SNAKECHARMER *scpacket)
{
    if (!memcmp(scpacket->magic, "NRT", 3) && scpacket->version == 1)
        return R_SUCCESS;
    else
        return R_FAIL;
}

