#include <uuid/uuid.h>
#include <mysql.h>
#include <syslog.h>

#include "config.h"
#include "rzb_intel_global.h"
#include "rzb_global.h"
#include "rzb_network.h"
#include "rzb_intel_listener.h"
#include "rzb_intel_db.h"
#include "rzb_management_log.h"
#include "rzb_component_global.h"
#include "rzb_component_db.h"

static HRESULT populateMailData(RZBNetworkConnection *net, unsigned size, char **dest);
static HRESULT receiveOptionalData(RZBNetworkConnection *net, char **data, unsigned size);
static HRESULT handleUpdateMailData(RZBNetworkConnection *net);

HRESULT handleIntel(RZBNetworkConnection *net)
{
    INTEL_TYPE intel_type;

    if (recvWrap(net, &intel_type, sizeof(intel_type)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to recieve mail_track message", "handleIntel");
        return R_FAIL;
    }

    switch (intel_type.msg_type)
    {
        case NEWMAIL:
            if ((handleNewMail(net)) != R_SUCCESS)
            {
                submitLog(LOG_ERR, "Failed to handle New Mail", "handleIntel");
                return R_FAIL;
            }
            break;
        case ADD_MAIL_DATA:
            if ((handleUpdateMailData(net)) != R_SUCCESS)
            {
                submitLog(LOG_ERR, "Failed to handle Mail Update", "handleIntel");
                return R_FAIL;
            }
            break;
        case ADD_MAIL_ATTACHMENT:
            if ((handleAddMailAttachment(net)) != R_SUCCESS)
            {
                submitLog(LOG_ERR, "Failed to handle add mail attachment", "handleIntel");
                return R_FAIL;
            }
            break;
        case NEWWEB:
            if ((handleWebTrack(net)) != R_SUCCESS)
            {
                submitLog(LOG_ERR, "Failed to handle web track data", "handleIntel");
                return R_FAIL;
            }
            break;
        case NEWDNS:
            if ((handleDNSTrack(net)) != R_SUCCESS)
            {
                submitLog(LOG_ERR, "Failed to handle dns track data", "handleIntel");
                return R_FAIL;
            }
            break;
        default:
            submitLog(LOG_ERR, "Incorrect packet type", "handleIntel");
            return R_FAIL;
    }
    return R_SUCCESS;
}

HRESULT handleDNSTrack(RZBNetworkConnection *net)
{
    SNAKECHARMER scpacket;
    DNS_TRACK dns_tracker;
    DNS_ENTRY dns_entry;
    HRESULT rval;

    if (recvWrap(net, &dns_tracker, sizeof(dns_tracker)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "failed to recieve dns_tracker data", "handleDNSTrack");
        return R_FAIL;
    }

    dns_entry.src_ip = dns_tracker.src_ip;
    dns_entry.dst_ip = dns_tracker.dst_ip;
    dns_entry.tx_id = dns_tracker.tx_id;
    dns_entry.flags = dns_tracker.flags;

    dns_entry.queries = NULL;
    dns_entry.answers = NULL;
    dns_entry.auth_ns = NULL;
    dns_entry.add_rec = NULL;

    rval = R_FAIL;
    do
    {
        if (receiveOptionalData(net, &dns_entry.queries, dns_tracker.queries_size) != R_SUCCESS)
            break;
        if (receiveOptionalData(net, &dns_entry.answers, dns_tracker.answers_size) != R_SUCCESS)
            break;
        if (receiveOptionalData(net, &dns_entry.auth_ns, dns_tracker.auth_ns_size) != R_SUCCESS)
            break;
        if (receiveOptionalData(net, &dns_entry.add_rec, dns_tracker.add_rec_size) != R_SUCCESS)
            break;
        if (insertDNSTrackData(&dns_entry, &dns_tracker) == R_SUCCESS)
            rval = R_SUCCESS;
    } while (0);
    acknowledge(net, &scpacket);
    return rval;
}

HRESULT handleWebTrack(RZBNetworkConnection *net)
{
    SNAKECHARMER scpacket;
    WEB_TRACK web_tracker;
    WEB_FINAL web_entry;
    HRESULT rval;

    if (recvWrap(net, &web_tracker, sizeof(WEB_TRACK)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to received web_tracker data", "handleWebTrack");
        return R_FAIL;
    }

    web_entry.src_ip = web_tracker.src_ip;
    web_entry.dst_ip = web_tracker.dst_ip;

    web_entry.url = NULL;
    web_entry.host = NULL;
    web_entry.user_agent = NULL;
    web_entry.cookie = NULL;
    web_entry.header = NULL;

    rval = R_FAIL;
    do
    {
        if (receiveOptionalData(net, &web_entry.url, web_tracker.url_size) != R_SUCCESS)
            break;
        if (receiveOptionalData(net, &web_entry.host, web_tracker.host_size) != R_SUCCESS)
            break;
        if (receiveOptionalData(net, &web_entry.user_agent, web_tracker.user_agent_size) != R_SUCCESS)
            break;
        if (receiveOptionalData(net, &web_entry.cookie, web_tracker.cookie_size) != R_SUCCESS)
            break;
        if (receiveOptionalData(net, &web_entry.header, web_tracker.header_size) != R_SUCCESS)
            break;
        if (insertWebTrackData(&web_entry, &web_tracker) == R_SUCCESS)
            rval = R_SUCCESS;
    } while (0);
    acknowledge(net, &scpacket);
    if (web_entry.url)
        free(web_entry.url);
    if (web_entry.host)
        free(web_entry.host);
    if (web_entry.user_agent)
        free(web_entry.user_agent);
    if (web_entry.cookie)
        free(web_entry.cookie);
    if (web_entry.header)
        free(web_entry.header);
    return rval;
}

static HRESULT receiveOptionalData(RZBNetworkConnection *net, char **data, unsigned size)
{
    char *tmp;

    *data = NULL;
    if (!size)
        return R_SUCCESS;

    if ((tmp = malloc(size)) == NULL)
    {
        submitLog(LOG_ERR, "Failed to allocate memory for data", "receiveOptionalData");
        return R_FAIL;
    }

    if (recvWrap(net, tmp, size) != R_SUCCESS)
    {
        free(tmp);
        submitLog(LOG_ERR, "Failed to receive additional web data", "receiveOptionalData");
        return R_FAIL;
    }

    *data = tmp;

    return R_SUCCESS;
}

HRESULT handleAddMailAttachment(RZBNetworkConnection *net)
{
    SNAKECHARMER scpacket;
    BASIC_COMPONENT base;
    ATTACHMENT_INFO attachment;
    COMPONENT component;
    char *name;

    if (recvWrap(net, &attachment, sizeof(ATTACHMENT_INFO)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to receive attachment information", "handleAddMailAttachment");
        return R_FAIL;
    }

    /* If the name size is not 0, the api will follow up by transmitting the name */
    if (attachment.name_size)
    {
        if ((name = malloc(attachment.name_size)) == NULL)
        {
            submitLog(LOG_ERR, "Failed to allocate memory for name", "handleAddMailAttachment");
            return R_FAIL;
        }
        if (recvWrap(net, name, attachment.name_size) != R_SUCCESS)
        {
            free(name);
            submitLog(LOG_ERR, "Failed to receive attachment name", "handleAddMailAttachment");
            return R_FAIL;
        }
    }
    else
        name = NULL;

    /* Populate the base component structure */
    base.size = attachment.size;
    memcpy(base.md5, attachment.md5, sizeof(base.md5));
    uuid_copy(base.datatype, attachment.data_type);

    if ((checkComponent(base.md5, base.size, &component)) != R_SUCCESS)
    {
        if (insertBaseComponent(&base) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to insert base component", "handleAddMailAttachment");
            if (name)
                free(name);
            return R_FAIL;
        }
    }
    else
        base.component_id = component.component_id;

    acknowledge(net, &scpacket);

    if (insertAttachment(attachment.mail_id, base.component_id, name) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed ot insert attachment data", "handleAddMailAttachment");
        if (name)
            free(name);
        return R_FAIL;
    }

    if (name)
        free(name);
    return R_SUCCESS;
}

static HRESULT handleUpdateMailData(RZBNetworkConnection *net)
{
    SNAKECHARMER scpacket;
    MAIL_UPDATE mupacket;
    const char *datatype;
    char *data;

    if (recvWrap(net, &mupacket, sizeof(mupacket)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to receive mail update packet", "handleUpdateMailData");
        return R_FAIL;
    }

    if ((data = malloc(mupacket.info_size)) == NULL)
    {
        submitLog(LOG_ERR, "Failed to allocate memory for data", "handleUpdateMailData");
        return R_FAIL;
    }

    if (recvWrap(net, data, mupacket.info_size) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to receive mail update packet", "handleUpdateMailData");
        free(data);
        return R_FAIL;
    }

    switch (mupacket.info_type)
    {
        case HEADER_DATE:
            datatype = "header_date";
            break;
        case MSG_ID:
            datatype = "message_id";
            break;
        case MIME_VER:
            datatype = "mime_ver";
            break;
        case LANGUAGE:
            datatype  = "content_language";
            break;
        case MAILER:
            datatype = "x_mailer";
            break;
        case THREAD_IDX:
            datatype = "thread_index";
            break;
        case BODY:
            datatype = "body";
            break;
        case XBLOCK:
            datatype = "x_blocks";
            break;
        default:
            submitLog(LOG_ERR, "Invalid mail update type", "handleUpdateMailData");
            free(data);
            return R_FAIL;
    }

    acknowledge(net, &scpacket);

    if ((insertMailUpdate(mupacket.mail_id, data, datatype)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to insert mail update", "handleUpdateMailData");
        free(data);
        return R_FAIL;
    }
    free(data);
    return R_SUCCESS;
}

HRESULT handleNewMail(RZBNetworkConnection *net)
{
    NEW_MAIL mail;
    NEW_MAIL_BLOBS blobs;
    NEW_MAIL_RESPONSE response;
    unsigned mail_id;

    if (recvWrap(net, &mail, sizeof(mail)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to receive new-mail message", "handleNewMail");
        return R_FAIL;
    }

    populateMailData(net, mail.mail_from_size, &blobs.mail_from);
    populateMailData(net, mail.rcpt_to_size, &blobs.rcpt_to);
    populateMailData(net, mail.subject_size, &blobs.subject);
    populateMailData(net, mail.msg_from_size, &blobs.msg_from);
    populateMailData(net, mail.msg_to_size, &blobs.msg_to);

    /* Populate mail_from */
    if ((insertNewMail(&blobs, &mail, &mail_id)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to insert new mail", "handleNewMail");
        return R_FAIL;
    }

    response.mail_id = mail_id;

    if (sendWrap(net, &response, sizeof(response)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send new mail response", "handleNewMail");
        return R_FAIL;
    }

    return R_SUCCESS;
}

static HRESULT populateMailData(RZBNetworkConnection *net, unsigned size, char **dest)
{
    if (size)
    {
        if ((*dest = malloc(size)) == NULL)
        {
            submitLog(LOG_ERR, "Failed tdo allocate memory for dest", "populateMailData");
            return R_MALLOC_FAIL;
        }

        if (recvWrap(net, *dest, size) != R_SUCCESS)
        {
            free(*dest);
            *dest = NULL;
            submitLog(LOG_ERR, "Failed to receive blob message", "populateMailData");
            return R_FAIL;
        }
    }
    else
        *dest = strdup(NULL_ENTRY);
    return R_SUCCESS;
}

