#include <uuid/uuid.h>
#include <syslog.h>
#include <arpa/inet.h>

#include "config.h"
#include "rzb_dispatcher.h"
#include "rzb_router.h"
#include "rzb_db.h"
#include "rzb_nug_manage.h"
#include "rzb_management_log.h"

typedef struct _UUIDLOOKUPTABLE
{
    uuid_t uuid;
    NUGGETENTRY *answer;
} UUIDLOOKUPTABLE;

static UUIDLOOKUPTABLE *typeTracker = NULL;
static UUIDLOOKUPTABLE *dataTracker = NULL;

static HRESULT addTypeTrack(NUGGETENTRY *entry);

HRESULT addNuggetEntry(NUGGETENTRY *entry)
{
    NUGGETENTRY *temp;
    printf("Adding nugget %s\n", entry->name);
    temp = entry;

    if (!(addTypeTrack(entry)) == R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to add type track data for new nuget", "addNuggetEntry");
        return R_FAIL;
    }

    if (!(addDataTrack(entry)) == R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to add data track data for new nugget", "addNuggetEntry");
        return R_FAIL;
    }

    if (!uuid_compare(entry->nugtype, DETECTION))
        addDefenseRouteEntry(temp);
    return R_SUCCESS;
}

static HRESULT addTypeTrack(NUGGETENTRY *entry)
{
    UUIDLOOKUPTABLE *cursor;
    NUGGETENTRY *curr;
    unsigned i;

    /* If necessary build the resolver table */
    cursor = typeTracker;
    if (!cursor)
    {
        submitLog(LOG_NOTICE, "Building App Entry Tracker", "addTypeTrack");
        cursor = calloc(validNuggetCount, sizeof(*cursor));
        typeTracker = cursor;
        if (!cursor)
        {
            perror("Failed to allocate memory for the Nugget Tracker Table\n");
            return R_MALLOC_FAIL;
        }

        for (i = 0; i < validNuggetCount; i++)
        {
            uuid_copy(cursor[i].uuid, validNuggetSet[i]);
            cursor[i].answer = NULL;
        }
    }
    /* Find the correct nugget type */
    for (i = 0, cursor = typeTracker; i < validNuggetCount; i++, cursor++)
    {
        if (!uuid_compare(cursor->uuid, entry->nugtype))
        {
            if (cursor->answer == NULL)
            {
                cursor->answer = entry;
            }
            else
            {
                curr = cursor->answer;
                while (curr->nextbytype != NULL)
                    curr = curr->nextbytype;
                curr->nextbytype = entry;
            }
            entry->nextbytype = NULL;
            return R_SUCCESS;
        }
    }
    submitLog(LOG_ERR, "Failed to add app entry", "addTypeTrack");
    return R_ADD_FAIL;
}

//#define DEBUG_NUG_MANAGEMENT    1
HRESULT addDataTrack(NUGGETENTRY *entry)
{
    UUIDLOOKUPTABLE *cursor;
    NUGGETENTRY *curr;
#ifdef DEBUG_NUG_MANAGEMENT
    char uuid_str1[SIZE_OF_UUID_STRING];
    char uuid_str2[SIZE_OF_UUID_STRING];
#endif
    unsigned i;

    /* If necessary build the resolver table */
    cursor = dataTracker;
    if (cursor == NULL)
    {
        submitLog(LOG_NOTICE, "Building Data Entry Tracker", "addTypeTrack");
        cursor = calloc(dataTypesCount, sizeof(*cursor));
        dataTracker = cursor;
        if (!cursor)
        {
            perror("Failed to allocate memory for the Nugget Tracker Table\n");
            return R_MALLOC_FAIL;
        }
        for (i = 0; i < dataTypesCount; i++)
        {
            uuid_copy(cursor[i].uuid, dataTypes[i]);
            cursor[i].answer = NULL;
        }
    }
    /* Find the correct nugget type */
#ifdef DEBUG_NUG_MANAGEMENT
    uuid_unparse(entry->datatype, uuid_str1);
    printf("Checking %s\n", uuid_str1);
#endif
    for (i = 0, cursor = dataTracker; i < dataTypesCount; i++, cursor++)
    {
#ifdef DEBUG_NUG_MANAGEMENT
        uuid_unparse(cursor->uuid, uuid_str2);
        printf("Against  %s\n", uuid_str2);
#endif
        if (!uuid_compare(cursor->uuid, entry->datatype))
        {
            if (!cursor->answer)
                cursor->answer = entry;
            else
            {
                curr = cursor->answer;
                while (curr->nextbydata)
                    curr = curr->nextbydata;
                curr->nextbydata = entry;
            }
            entry->nextbydata = NULL;
            //printDataTable();
            return R_SUCCESS;
        }
    }
    submitLog(LOG_ERR, "Failed to add Data entry", "addDataTrack");
    return R_ADD_FAIL;
}

void printTypeTable(void)
{
    unsigned i = 0;
    UUIDLOOKUPTABLE *cursor = typeTracker;
    NUGGETENTRY *entry;
    char dataName[MAX_DESCRIPTION_SIZE];
    char appName[MAX_DESCRIPTION_SIZE];
    char uuid_str[SIZE_OF_UUID_STRING];
    char ip_str[INET6_ADDRSTRLEN];

    printf("\nType Track Table\n");
    printf("================\n");

    if (typeTracker == NULL)
    {
        printf("Table is NULL\n");
        return;
    }

    for (i=0; i < validNuggetCount; i++)
    {
//		printf("\nCycle: %u %d\n", i, sizeof(UUIDLOOKUPTABLE));
        uuid_unparse(cursor[i].uuid, uuid_str);
        getUUIDDesc(uuid_str, dataName);
        printf("%s (%s):\n", dataName, uuid_str);
        entry = cursor[i].answer;
        if (entry == NULL)
            printf("\tNO REGISTERED NUGGETS\n");
        while (entry != NULL)
        {
            uuid_unparse(entry->nugapp, uuid_str);
            getUUIDDesc(uuid_str, appName);
            printf("\tID: %u, Name: %s, App: %s, IP: %s, Port: %u\n", entry->nuggetid, entry->name, appName, inet_ntop(AF_INET, &entry->nugaddr, ip_str, sizeof(ip_str)), (unsigned)entry->nugport);
            entry = entry->nextbytype;
        }
    }
}

void printDataTable(void)
{
    unsigned i = 0;
    UUIDLOOKUPTABLE *cursor = dataTracker;
    NUGGETENTRY *entry;
    char dataName[MAX_DESCRIPTION_SIZE];
    char appName[MAX_DESCRIPTION_SIZE];
    char uuid_str[SIZE_OF_UUID_STRING];
    char ip_str[INET6_ADDRSTRLEN];

    printf("\nData Track Table\n");
    printf("================\n");

    if (dataTracker == NULL)
    {
        printf("Table is NULL\n");
        return;
    }

    for (i = 0; i < dataTypesCount; i++)
    {
//		printf("\nCycle: %u\n", i);
        uuid_unparse(cursor->uuid, uuid_str);
        getUUIDDesc(uuid_str, dataName);
        printf("%s (%s):\n", dataName, uuid_str);
        entry = cursor->answer;
        if (entry == NULL)
            printf("\tNO REGISTERED NUGGETS\n");
        while (entry != NULL)
        {
            uuid_unparse(entry->nugapp, uuid_str);
            getUUIDDesc(uuid_str, appName);
            printf("\tID: %u, Name: %s, App: %s, IP: %s, Port: %u\n", entry->nuggetid, entry->name, appName, inet_ntop(AF_INET, &entry->nugaddr, ip_str, sizeof(ip_str)), (unsigned)entry->nugport);
            entry = entry->nextbydata;
        }
        cursor++;
    }
}

HRESULT deleteNuggetEntry(unsigned nugid)
{
    unsigned i = 0;
    UUIDLOOKUPTABLE *cursor = dataTracker;
    NUGGETENTRY *entry;
    NUGGETENTRY **pentry;
    NUGGETENTRY *target = NULL;

    if (dataTracker == NULL || typeTracker == NULL)
        return R_FAIL;

    /* Clear out the data type table */

    for (i=0; i < dataTypesCount; i++)
    {
        for (pentry = &cursor->answer; (entry = *pentry); pentry = &entry->nextbydata)
        {
            if (entry->nuggetid == nugid)
            {
                *pentry = entry->nextbydata;
                target = entry;
            }
        }
        cursor++;
    }

    /* Clear out the nugget type table */
    cursor = typeTracker;
    for (i=0; i < validNuggetCount; i++)
    {
        for (pentry = &cursor->answer; (entry = *pentry); pentry = &entry->nextbytype)
        {
            if (entry->nuggetid == nugid)
            {
                *pentry = entry->nextbydata;
                target = entry;
            }
        }
        cursor++;
    }

    if (target)
        free(target);

    return R_SUCCESS;
}

HRESULT findByID(NUGGETENTRY **entry, unsigned nugid, uuid_t type)
{
    UUIDLOOKUPTABLE *cursor;
    NUGGETENTRY *current;
    unsigned i;
    cursor = typeTracker;

    if (cursor == NULL)
        return R_FAIL;

    for (i=0; i<validNuggetCount; i++)
    {
        if ((uuid_compare(cursor->uuid, type)) == 0)
        {
            current = cursor->answer;
            while (current != NULL)
            {
                if (current->nuggetid == nugid)
                {
                    *entry = current;
                    return R_FOUND;
                }
                current = current->nextbytype;
            }
        }
        cursor++;
    }
    return R_NOT_FOUND;
}

