#include <arpa/inet.h>

#include "config.h"
#include "rzb_global.h"
#include "rzb_router.h"
#include "rzb_db.h"
#include "rzb_nug_manage.h"

static DATAENTRY *routingTable = NULL;

DATAENTRY *getRoutingTableLoc(void)
{
    return routingTable;
}

/*
 * printRoutingTable
 * input:  none
 * output: HRESULT
 * author: Matthew Olney
 *
 * This function is primarily for debugging.  It prints the defense routing table.
 * For each data type, it prints each nugget application type that has been
 * registered to it, along with each entry that is associated with that application type.
*/

HRESULT printRoutingTable(void)
{
    DATAENTRY *de;
    APPENTRY *ae;
    ROUTEENTRY *re;
    NUGGETENTRY *ne;
    char desc[MAX_DESCRIPTION_SIZE];
    char uuid_str[SIZE_OF_UUID_STRING];
    char ip_str[INET6_ADDRSTRLEN];

    printf("\nDefense Routing Table\n");
    printf(  "=====================\n");
    for (de = routingTable; de; de = de->next)
    {
        uuid_unparse(de->datatype, uuid_str);
        getUUIDDesc(uuid_str, desc);
        printf("Data Type: %s (%s)\n", desc, uuid_str);
        for (ae = de->apphead; ae; ae = ae->next)
        {
            uuid_unparse(ae->nugapp, uuid_str);
            getUUIDDesc(uuid_str, desc);
            printf("\tApp Type: %s (%s)\n", desc, uuid_str);
            for (re = ae->routehead; re; re = re->next)
            {
                ne = re->nugget;
                printf("\t\tNugget ID: %u\n", ne->nuggetid);
                printf("\t\t\tName: %s\n", ne->name);
                printf("\t\t\tSocket: %s:%u\n", inet_ntop(AF_INET, &ne->nugaddr, ip_str, sizeof(ip_str)), (unsigned)ne->nugport);
                printf("\t\t\tNugget Mem Loc: %p\n", (void *)re->nugget);
            }
        }
    }

    return R_SUCCESS;
}

/*
 * addRoute
 * input:  APPENTRY (passed by reference), NUGGETENTRY (passed by reference)
 * output: HRESULT
 * author: Matthew Olney
 *
 * This function adds an additional entry into the list of nuggets that have
 * been registered to a specific nugget application type.
*/

static HRESULT addRoute(APPENTRY *app, NUGGETENTRY *entry)
{
    ROUTEENTRY **tail_re;
    ROUTEENTRY *re;

    if ((re = calloc(1, sizeof(*re))) == NULL)
    {
        perror("Failed to malloc for routing table build - Route Entry\n");
        return R_MALLOC_FAIL;
    }

    re->nugget = entry;

    /* Find the end of the data list */
    for (tail_re = &app->routehead; *tail_re; tail_re = &(*tail_re)->next);
    *tail_re = re;

    if (!app->ondeck)
        app->ondeck = re;

    return R_SUCCESS;
}

/*
 * addAppEntry
 * input:  DATAENTRY (passed by reference), NUGGETENTRY (passed by reference)
 * output: HRESULT
 * author: Matthew Olney
 *
 * This function adds a nugget application entry to an existing data type.
*/

static HRESULT addAppEntry(DATAENTRY *de, NUGGETENTRY *entry)
{
    APPENTRY **tail_ae;
    APPENTRY *ae;

    if ((ae = calloc(1, sizeof(*ae))) == NULL)
    {
        perror("Failed to malloc for routing table build - Application Entry\n");
        return R_MALLOC_FAIL;
    }

    /* Add the app data and populate struct */
    uuid_copy(ae->nugapp, entry->nugapp);

    /* Find the end of the data list */
    for (tail_ae = &de->apphead; *tail_ae; tail_ae = &(*tail_ae)->next);
    *tail_ae = ae;

    return addRoute(ae, entry);
}


/*
 * addDataEntry
 * input: NUGGETENTRY (passed by reference)
 * output: HRESULT
 * author: Matthew Olney
 *
 * This function adds a new data entry into the routing table.
*/

static HRESULT addDataEntry(NUGGETENTRY *entry)
{
    DATAENTRY **tail_de;
    DATAENTRY *de;

    if ((de = calloc(1, sizeof(*de))) == NULL)
    {
        perror("Failed to allocate memory for addDataEntry\n");
        return R_MALLOC_FAIL;
    }

    /* Populate the dataentry */
    uuid_copy(de->datatype, entry->datatype);

    /* Find the end of the data list */
    for (tail_de = &routingTable; *tail_de; tail_de = &(*tail_de)->next);
    *tail_de = de;

    return addAppEntry(de, entry);
}

/*
 * removeRouteByNugID
 * input: unsigned nugid
 * ouptut: HRESULT
 * author: Matthew Olney
 *
 * This function traverses the Routing Table and removes all entries for a given NugID.
*/

HRESULT removeRouteByNugID(unsigned nugid)
{
    DATAENTRY *de;
    APPENTRY *ae;
    ROUTEENTRY *re;
    ROUTEENTRY **i_re;
    NUGGETENTRY *ne;

    if (routingTable == NULL)
        return R_SUCCESS;

    printf("Removing route entry for %u\n", nugid);

    /* Loop through all Data entries */
    for (de = routingTable; de; de = de->next)
    {
        for (ae = de->apphead; ae; ae = ae->next)
        {
            for (i_re = &ae->routehead; (re = *i_re); )
            {
                ne = re->nugget;
                /* Found a route entry with the correct ID */
                if (ne->nuggetid == nugid)
                {
                    printf("Found route entry to remove for %u\n", nugid);
                    /* Remove from the list */
                    *i_re = re->next;
                    /* Update ondeck, if it is being removed */
                    if (ae->ondeck == re)
                    {
                        if (re->next)
                            ae->ondeck = re->next;
                        else
                            ae->ondeck = ae->routehead;
                    }
                    free(re);
                }
                else
                    i_re = &(*i_re)->next;
            }
        }
    }
    return R_SUCCESS;
}

/*
 * findDataEntry
 * input:  DATAENTRY (passed by reference), NUGGETENTRY (passed by reference)
 * output: HRESULT, populates **found with the memory location of the desired data entry in the routing table
 * author: Matthew Olney
 *
 * This function finds the entry in the routing table to the correspondign data type.  If it finds it,
 * it sets the found poitner to the location in the table and returns R_FOUND.  If it does not find it,
 * it returns R_NOT_FOUND
*/

HRESULT findDataEntry(DATAENTRY **found, NUGGETENTRY *entry)
{
    DATAENTRY *de;

    for (de = routingTable; de; de = de->next)
    {
        if (!(uuid_compare(de->datatype, entry->datatype)))
        {
            *found = de;
            return R_FOUND;
        }
    }

    // Not found.  Set return pointer to NULL
    *found = NULL;
    return R_NOT_FOUND;
}

/*
 * findAppEntry
 * input:  DATAENTRY (passed by reference), APPENTRY (passed by reference), NUGGETENTRY (passed by reference)
 * output: HRESULT, Populates **found with a pointer to the appentry in the routing table
 * author: Matthew Olney
 *
 * This table finds the entry in the routing table that corresponds to the nugget entry's app type
 * If it fails to find an entry, it returns R_NOT_FOUND.  If it does find an entry, it copies
 * a pointer to that locaiton into the *found variable and returns R_FOUND.
*/

static HRESULT findAppEntry(DATAENTRY *de, APPENTRY **found, NUGGETENTRY *entry)
{
    APPENTRY *ae;

    for (ae = de->apphead; ae; ae = ae->next)
    {
        if (!(uuid_compare(ae->nugapp, entry->nugapp)))
        {
            *found = ae;
            return R_FOUND;
        }
    }

    // Not found.  Set return pointer to NULL
    *found = NULL;
    return R_NOT_FOUND;
}

/*
 * addDefenseRouteEntry
 * input:  NUGGETENTRY (passed by reference)
 * output: HRESULT
 * author: Matthew Olney
 *
 * This function is called when a nugget is registered that is declared as
 * a detection nugget.  It determines if the routing table has been built,
 * if the data type has been declared and if the app type has been previously
 * registered and calls the appropriate function.
*/

HRESULT addDefenseRouteEntry(NUGGETENTRY *entry)
{
    DATAENTRY *curr_de;
    APPENTRY *curr_ae;

    if ((findDataEntry(&curr_de, entry)) == R_FOUND)
    {
        if ((findAppEntry(curr_de, &curr_ae, entry)) == R_FOUND)
            addRoute(curr_ae, entry);
        else
            addAppEntry(curr_de, entry);
    }
    else
        addDataEntry(entry);

    return R_SUCCESS;
}

