#include <syslog.h>

#include "config.h"
#include "rzb_global.h"
#include "rzb_db.h"
#include "rzb_router.h"
#include "rzb_network.h"
#include "rzb_management_log.h"
#include "rzb_workstation_db.h"
#include "rzb_workstation_listener.h"
#include "rzb_workstation_global.h"
#include "rzb_nug_manage.h"

HRESULT handleWorkstation(RZBNetworkConnection *net)
{
    WKST_REQ_TYPE request_type;

    if (recvWrap(net, &request_type, sizeof(request_type)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to receive workstation request message", "handleWorkstation");
        return R_FAIL;
    }

    switch (request_type.msg_type)
    {
        case SHOW_ROUTING_TABLE:
            if ((handleShowRoutingTable(net)) != R_SUCCESS)
            {
                submitLog(LOG_ERR, "Failed to handle show routing table", "handleWorkstation");
                return R_FAIL;
            }
            break;
        case SHOW_LOGS:
            if ((handleShowLog(net)) != R_SUCCESS)
            {
                submitLog(LOG_ERR, "Failed to handle show logs", "handleWorkstation");
                return R_FAIL;
            }
            break;
        default:
            submitLog(LOG_ERR, "Invalid workstation request type", "handleWorkstation");
            return R_FAIL;
    }

    return R_SUCCESS;
}

HRESULT handleShowLog(RZBNetworkConnection *net)
{
    LOG_REQUEST_PACKET req;

    if (recvWrap(net, &req, sizeof(req)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Unable to receive log request packet", "handleShowLog");
        return R_FAIL;
    }

    switch (req.type)
    {
        case LOGS_BY_NUMBER:
            if (fetchLogsByNumber(net, req.data) != R_SUCCESS)
            {
                submitLog(LOG_ERR, "Failed to fetech logs by number", "handleShowLog");
                return R_FAIL;
            }
            break;
        default:
            submitLog(LOG_ERR, "Invalid log request", "handleShowLog");
            return R_FAIL;
    }

    return R_SUCCESS;
}

HRESULT handleShowRoutingTable(RZBNetworkConnection *net)
{
    WKST_DATA_RESPONSE response;
    DATAENTRY *route_table = getRoutingTableLoc();

    /* If route_table is NULL, no detection nuggets have registered yet */
    if (route_table == NULL)
    {
        response.resp_type = NO_ROUTE_TABLE;
        response.resp_size = 0;
        sendWrap(net, &response, sizeof(response));
        return R_SUCCESS;
    }

    /* Start the process of sending the routing table */
    response.resp_type = ROUTE_TABLE_START;
    response.resp_size = 0;

    if (sendWrap(net, &response, sizeof(response)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send response packet", "handleShowRoutingTable");
        return R_FAIL;
    }

    while (route_table != NULL)
    {
        if (sendDataRouteEntry(&route_table, &response, net) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to send data route entry", "handleShowRoutingTable");
            return R_FAIL;
        }
    }

    response.resp_type = ROUTE_TABLE_END;

    if (sendWrap(net, &response, sizeof(response)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send response packet", "handleShowRoutingTable");
        return R_FAIL;
    }

    return R_SUCCESS;
}

HRESULT sendDataRouteEntry(DATAENTRY **data_cursor, WKST_DATA_RESPONSE *response, RZBNetworkConnection *net)
{
    APPENTRY *app_cursor;
    DATAENTRY *dc = *data_cursor;
    ROUTE_HEADER_PACKET entry;
    char uuid_str[SIZE_OF_UUID_STRING];
    char desc[MAX_DESCRIPTION_SIZE];

    /* Announce the response type */
    response->resp_type = ROUTE_DATA_ENTRY;
    response->resp_size = 0;
    if (sendWrap(net, response, sizeof(*response)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send response packet", "handleShowRoutingTable");
        return R_FAIL;
    }

    /* Build the headere entry packet */
    uuid_unparse(dc->datatype, uuid_str);
    getUUIDDesc(uuid_str, desc);
    uuid_copy(entry.datatype, dc->datatype);
    strncpy(entry.dt_name, desc, MAX_DESCRIPTION_SIZE);

    if (sendWrap(net, &entry, sizeof(entry)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Unable to send data type entry", "sendDataRouteEntry");
        return R_FAIL;
    }

    app_cursor = dc->apphead;

    while (app_cursor != NULL)
    {
        if (sendAppRouteEntry(&app_cursor, response, net) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to send app route data", "sendDataRouteEntry");
            return R_FAIL;
        }
    }

    *data_cursor = dc->next;
    return R_SUCCESS;
}

HRESULT sendAppRouteEntry(APPENTRY **app_cursor, WKST_DATA_RESPONSE *response, RZBNetworkConnection *net)
{
    ROUTEENTRY *entry_cursor;
    ROUTE_HEADER_PACKET entry;
    APPENTRY *ae = *app_cursor;
    char uuid_str[SIZE_OF_UUID_STRING];
    char desc[MAX_DESCRIPTION_SIZE];

    /* Announce the response type */
    response->resp_type = ROUTE_APP_ENTRY;
    response->resp_size = 0;
    if (sendWrap(net, response, sizeof(*response)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send response packet", "handleShowRoutingTable");
        return R_FAIL;
    }

    /* Build the headere entry packet */
    uuid_unparse(ae->nugapp, uuid_str);
    getUUIDDesc(uuid_str, desc);
    uuid_copy(entry.datatype, ae->nugapp);
    strncpy(entry.dt_name, desc, MAX_DESCRIPTION_SIZE);

    if (sendWrap(net, &entry, sizeof(entry)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Unable to send data type entry", "sendAppRouteEntry");
        return R_FAIL;
    }

    entry_cursor = ae->routehead;
    while (entry_cursor != NULL)
    {
        if (sendNugEntry(&entry_cursor, response, net) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to send nugget entry data", "sendAppRouteEntry");
            return R_FAIL;
        }
    }

    *app_cursor = ae->next;
    return R_SUCCESS;
}

HRESULT sendNugEntry(ROUTEENTRY **entry_cursor, WKST_DATA_RESPONSE *response, RZBNetworkConnection *net)
{
    ROUTE_NUGGET_PACKET entry;
    ROUTEENTRY *re = *entry_cursor;
    NUGGETENTRY *nug_data = re->nugget;

    response->resp_type = ROUTE_NUGGET_ENTRY;
    response->resp_size = 0;
    if (sendWrap(net, response, sizeof(*response)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Failed to send response packet", "handleShowRoutingTable");
        return R_FAIL;
    }

    entry.nugid = nug_data->nuggetid;
    entry.nugaddr.s_addr = nug_data->nugaddr.s_addr;
    entry.nugport = nug_data->nugport;
    strncpy(entry.name, nug_data->name, sizeof(entry.name));

    if (sendWrap(net, &entry, sizeof(entry)) != R_SUCCESS)
    {
        submitLog(LOG_ERR, "Unable teo send nugget entry data", "sendNugEntry");
        return R_FAIL;
    }
    *entry_cursor = re->next;

    return R_SUCCESS;
}

