#include "config.h"
#include "swf_scanner.h"

/* Global Variables */
const DetectionAPI *gDetectionAPI;

static void sendWarnings(ALERT *alert, ErrorCode errCode);

#ifdef VERBOSE

static void printRECT(RECT *struct_rect, char *prefix)
{
    printf("%sRECT.NBits    [%u]\n", prefix, struct_rect->Nbits);
    printf("%sRECT.Xmin     [%u]\n", prefix, struct_rect->Xmin);
    printf("%sRECT.Xmax     [%u]\n", prefix, struct_rect->Xmax);
    printf("%sRECT.Ymin     [%u]\n", prefix, struct_rect->Ymin);
    printf("%sRECT.Ymax     [%u]\n", prefix, struct_rect->Ymax);
}

static void printHEADER(SWFFileInfo *swffile)
{
    printf("SWF.HEADER.Signature [%c%c%c]\n", swffile->header.Signature[0], swffile->header.Signature[1], swffile->header.Signature[2]);
    printf("SWF.HEADER.Version   [%d]\n", swffile->header.Version);
    printf("SWF.HEADER.FileLength [%d]\n", swffile->header.FileLength);
    printf("SWF.HEADER.FrameSize\n");
    printRECT(&(swffile->header.FrameSize), "-- ");
    printf("SWF.HEADER.FrameRate [%d]\n", swffile->header.FrameRate);
    printf("SWF.HEADER.FrameCount [%d]\n", swffile->header.FrameCount);
}

static void printTAGs(TAGList *tags, uint32_t level)
{
    TAG *cur_tag;
    uint16_t tag_type;
    uint32_t tag_index = 0;

    char *prefix;

    if (level == 0)
        prefix = "";
    else
        prefix = "--- ";

    cur_tag = tags->tag_head;

    while (cur_tag != NULL)
    {
        tag_type = cur_tag->TagCode;

        if (tag_type < MAX_TYPE_CODE)
            printf("%sTAG[%d].TagCode [%d][%s]\n", prefix, tag_index, tag_type, TypeName[tag_type]);
        else
            printf("%sTAG[%d].TagCode [%d]\n", prefix, tag_index, tag_type);

        printf("%sTAG[%d].Length [%d]\n", prefix, tag_index, cur_tag->Length);

        cur_tag = cur_tag->next;
        tag_index++;

        if (tag_type == TYPE_DefineSprite)
            printTAGs(&(cur_tag->tags), 1);
    }
}

#endif

/* Decompression function for 'CWS' Flash files. It decompress an entire file at once */

#define CHUNK 0x40000  // 256K

static ErrorCode decompressSWF(SWFFileInfo *swffile, uint8_t *cursor, uint32_t size)
{
    z_stream stream;
    int ret;
    int count_newbuff = 0;
    uint32_t file_length;

    stream.next_in   = cursor + 8;
    stream.avail_in  = size - 8;

    stream.zalloc    = Z_NULL;
    stream.zfree     = Z_NULL;
    stream.opaque    = Z_NULL;

    stream.avail_out = 0;
    inflateInit(&stream);

    swffile->decompressed_data = (uint8_t *)malloc(CHUNK);
    if (swffile->decompressed_data == NULL)
        return SWF_ERROR_MEMORY;

    count_newbuff++;
    memcpy(swffile->decompressed_data, &(swffile->header), 4);

    // The first 8 bytes of SWF file is not compressed
    stream.avail_out = CHUNK - 8;
    stream.next_out  = swffile->decompressed_data + 8;

    while (1)
    {
        if (!stream.avail_out)
        {
            count_newbuff++;
            swffile->decompressed_data = (Bytef *)realloc(swffile->decompressed_data, CHUNK * count_newbuff);
            if (swffile->decompressed_data)
                return SWF_ERROR_MEMORY;

            stream.next_out  = swffile->decompressed_data + ((count_newbuff - 1) * CHUNK);
            stream.avail_out = CHUNK;
        }

        if (stream.next_in)
            ret = inflate(&stream, Z_NO_FLUSH);
        else
            ret = inflate(&stream, Z_FINISH);

        if (ret != Z_OK && ret != Z_STREAM_END)
            break;

        if (ret == Z_STREAM_END)
        {
            ret = inflateEnd(&stream);

            file_length = stream.total_out + 8;   // 8 bytes = header.Signature, header.Version, header.FileLength

            // store the size of decompressed data to the file length field in the buffer
            *(swffile->decompressed_data + 4) = (uint8_t)file_length;
            *(swffile->decompressed_data + 5) = (uint8_t)(file_length >> 8);
            *(swffile->decompressed_data + 6) = (uint8_t)(file_length >> 16);
            *(swffile->decompressed_data + 7) = (uint8_t)(file_length >> 24);

            // use SWF scanner's own memory instead of the memory provided by the dispatcher
            swffile->bof = swffile->decompressed_data;
            swffile->eof = swffile->decompressed_data + file_length;

            break;
        }
    }

    if (ret != Z_OK)
    {
        free(swffile->decompressed_data);
        swffile->decompressed_data = NULL;
        swffile->compressed = 0;

        return SWF_ERROR_DECOMPRESSION;
    }

    return SWF_OK;
}

static ErrorCode createTAG(TAGList *tag_list)
{
    TAG *new_tag;

    new_tag = (TAG *)malloc(sizeof(TAG));

    if (new_tag == NULL)
        return SWF_ERROR_MEMORY;

    memset(new_tag, 0, sizeof(TAG));

    if (tag_list->tag_head == NULL)
    {
        tag_list->tag_head = new_tag;
        tag_list->tag_tail = new_tag;
    }
    else
    {
        ((TAG *)tag_list->tag_tail)->next = new_tag;
        new_tag->prev            = tag_list->tag_tail;

        tag_list->tag_tail = new_tag;
    }

    return SWF_OK;
}

static void freeTAGs(TAGList *tags)
{
    TAG *tag_head;
    TAG *tag_delete;

    tag_head = tags->tag_head;

    while (tag_head != NULL)
    {
        tag_delete = tag_head;
        tag_head   = tag_head->next;

        tag_delete->prev = NULL;
        tag_delete->next = NULL;

        if (tag_delete->TagCode == TYPE_DefineSprite)
            freeTAGs(&(tag_delete->tags));

        free(tag_delete);
    }

    tags->tag_head = NULL;
    tags->tag_tail = NULL;
}

// Make sure Nbits <= 31

static uint32_t readVarBits(uint8_t **cursor, uint8_t Nbits, uint8_t *bit_offset)
{
    uint32_t multibits_value;    // the most significant bit is the leftmost bit

    uint8_t  i;
    uint8_t *tmp_cursor = *cursor;
    uint8_t  tmp_offset = *bit_offset;

    multibits_value = 0;
    for (i = 0; i < Nbits; i++)
    {
        if (*tmp_cursor & (0x80 >> (tmp_offset % 8)))
            multibits_value |= 1;

        multibits_value <<= 1;
        tmp_offset++;

        if ((tmp_offset % 8) == 0)
            tmp_cursor++;
    }

    multibits_value >>= 1; // to cancel the last left shift

    *cursor = tmp_cursor;
    *bit_offset = tmp_offset;

    return multibits_value;
}

static ErrorCode parseRECT(SWFFileInfo *swffile, uint8_t **cursor)
{
    uint8_t Nbits;
    uint8_t bit_offset;
    uint16_t rect_size;

    uint8_t *tmp_cursor = *cursor;

    if (tmp_cursor + 1 > swffile->eof)
        return SWF_ERROR_NODATA;

    Nbits = (uint8_t)((*tmp_cursor & 0xF8) >> 3);  // read left most 5 bits
    rect_size = (uint16_t)ceil((double)(5 + (Nbits * 4))/8) ;   // total in bytes

    if (tmp_cursor + rect_size > swffile->eof)
        return SWF_ERROR_NODATA;

    swffile->header.FrameSize.Nbits = Nbits;

    // after Nbits field
    bit_offset = 5;
    swffile->header.FrameSize.Xmin = readVarBits(&tmp_cursor, Nbits, &bit_offset);
    swffile->header.FrameSize.Xmax = readVarBits(&tmp_cursor, Nbits, &bit_offset);
    swffile->header.FrameSize.Ymin = readVarBits(&tmp_cursor, Nbits, &bit_offset);
    swffile->header.FrameSize.Ymax = readVarBits(&tmp_cursor, Nbits, &bit_offset);

    *cursor += rect_size;

    return SWF_OK;
}

static ErrorCode parseHEADER(SWFFileInfo *swffile, uint8_t **cursor)
{
    uint8_t *tmp_cursor = *cursor;

    if (tmp_cursor + 12 > swffile->eof)
    {
        DEBUG_SWF(printf("Not enough data to parse SWF header\n");)
        return SWF_ERROR_NODATA;
    }

    tmp_cursor += 4;
    swffile->header.FileLength = READ_LITTLE_32(tmp_cursor);
    tmp_cursor += 4;

#if 0
    // File length does not follow the specification all the time
    if (swffile->header.FileLength != (swffile->eof - swffile->bof))
    {
        DEBUG_SWF(printf("swffile->header.FileLength [0x%08x] actual [0x%08x]\n",
                         swffile->header.FileLength, (swffile->eof - swffile->bof));)
        return SWF_ERROR_INVALID_FILELENGTH;
    }
#endif

    parseRECT(swffile, &tmp_cursor);

    if (tmp_cursor + 4 > swffile->eof)
        return SWF_ERROR_NODATA;

    swffile->header.FrameRate  = READ_LITTLE_16(tmp_cursor);
    tmp_cursor += 2;
    swffile->header.FrameCount = READ_LITTLE_16(tmp_cursor);
    tmp_cursor += 2;

    *cursor = tmp_cursor;

    VERBOSE_SWF(printHEADER(swffile);)

    return SWF_OK;
}

static ErrorCode parseTAG(TAGList *tag_list, uint8_t **cursor, uint8_t *end_to_parse)
{
    uint8_t  *tmp_cursor = *cursor;
    uint16_t  type_and_length;
    uint16_t  tag_type;
    uint32_t  tag_length;
    ErrorCode errCode;

    if (tmp_cursor + 2 > end_to_parse)
        return SWF_ERROR_NODATA;

    type_and_length = READ_LITTLE_16(tmp_cursor);
    tmp_cursor += 2;

    tag_type   = (uint16_t)((type_and_length & 0xFFC0) >> 6);
    tag_length = type_and_length & 0x003F;

    if (tag_length == 0x3F)
    {
        if (tmp_cursor + 4 > end_to_parse)
            return SWF_ERROR_NODATA;

        tag_length = READ_LITTLE_32(tmp_cursor);
        tmp_cursor += 4;
    }

    if (tmp_cursor + tag_length > end_to_parse)
        return SWF_ERROR_NODATA;

    if (tmp_cursor + tag_length < tmp_cursor) // integer overflow
        return SWF_ERROR_INVALID_TAGLENGTH;

    // Once a tag structure is allocated, there must be tag_length data
    errCode = createTAG(tag_list);
    if (errCode != SWF_OK)
        return errCode;

    ((TAG *)tag_list->tag_tail)->TagCode   = tag_type;
    ((TAG *)tag_list->tag_tail)->Length    = tag_length;
    ((TAG *)tag_list->tag_tail)->tag_start = *cursor;
    ((TAG *)tag_list->tag_tail)->tag_data  = tmp_cursor;

    *cursor = tmp_cursor + tag_length;

    return SWF_OK;
}

static ErrorCode parseDefineSpriteTAG(TAG *definesprite_tag)
{
    ErrorCode errCode    = SWF_OK;
    uint8_t  *tag_cursor = definesprite_tag->tag_data;
    uint8_t  *end_of_tag = tag_cursor + definesprite_tag->Length;

    // In order to include tag array,
    // the length of DefineSprite tag should be bigger than or equal to 6
    if (definesprite_tag->Length < 6)
        return SWF_ERROR_INVALID_TAGLENGTH;

    tag_cursor += 4; // to skip over Sprite ID and FrameCount fields

    while (errCode == SWF_OK && tag_cursor < end_of_tag)
        errCode = parseTAG(&(definesprite_tag->tags), &tag_cursor, end_of_tag);

    return errCode;
}

static ErrorCode parseSWF(SWFFileInfo *swffile, uint8_t *cursor, uint32_t size, ALERT *alert)
{
    ErrorCode errCode;

    if ((*cursor == 'F' || *cursor == 'C')
        && *(cursor + 1) == 'W' && *(cursor + 2) == 'S')
    {
        swffile->header.Signature[0] = 'F';
        swffile->header.Signature[1] = 'W';
        swffile->header.Signature[2] = 'S';

        if (*cursor == 'C')
            swffile->compressed = 1;
    }
    else
        return SWF_ERROR_INVALID_SIG;

    swffile->header.Version = *(cursor + 3);

    if (swffile->compressed)
    {
        if ((errCode = decompressSWF(swffile, cursor, size)) == SWF_OK)
            cursor = swffile->decompressed_data;
        else
            return SWF_ERROR_DECOMPRESSION;
    }

    errCode = parseHEADER(swffile, &cursor);

    // parseHEADER returns either SWF_OK or SWF_ERROR_NODATA
    // We don't need to parse further in this case if SWF_ERROR_NODATA takes place in the header
    if (errCode != SWF_OK)
        return errCode;

    while (errCode == SWF_OK && cursor < swffile->eof)
    {
        errCode = parseTAG(&(swffile->tags), &cursor, swffile->eof);

        if (errCode == SWF_OK && ((TAG *)swffile->tags.tag_tail)->TagCode == TYPE_DefineSprite)
            errCode = parseDefineSpriteTAG(swffile->tags.tag_tail);
    }

    // One of tags can have wrong tag length, we want to handle as many as tags possible
    if (errCode != SWF_OK)
        sendWarnings(alert, errCode);

    VERBOSE_SWF(printTAGs(&(swffile->tags), 0);)

    return SWF_OK;
}

static ErrorCode scanActionRecords(TAG *action_tag, SWFDetectionID *detectID)
{
    uint8_t *action_cursor = action_tag->tag_data;
    uint8_t *end_of_tag    = action_tag->tag_data + action_tag->Length;

    uint8_t  action_code;
    uint16_t action_length;

    uint8_t *detect_cursor;

    *detectID = SWFDetectionID_NONE;

    while (action_cursor < end_of_tag)
    {
        action_code = *action_cursor;
        action_cursor++;

        if (action_code < 0x80)
            continue;
        else
        {
            if (action_cursor + 2 > end_of_tag)
                return SWF_ERROR_NODATA;

            action_length = READ_LITTLE_16(action_cursor);
            action_cursor += 2;

            if (action_code == ACTIONCODE_ActionDefineFunction || action_code == ACTIONCODE_ActionDefineFunction2)
            {
                detect_cursor = action_cursor;

                while (*detect_cursor != '\0') detect_cursor++;

                if (detect_cursor >= action_cursor + action_length)
                {
                    *detectID = SWFDetectionID_CVE_2005_2628;
                    return SWF_OK;
                }
            }

            action_cursor += action_length;
        }
    }

    return SWF_OK;
}

static ErrorCode scanTAGs(TAGList *tags, SWFDetectionID *detectID, uint16_t frame_count)
{
    TAG      *cur_tag;
    uint8_t  *detect_cursor;
    uint8_t  *tmp_cursor;
    ErrorCode sub_error;

    uint32_t  endtag_found    = 0;
    uint32_t  count_showframe = 0;  // for CVE-2007-5400

    uint16_t  definesprite_framecount;

    *detectID = SWFDetectionID_NONE;
    cur_tag   = tags->tag_head;

    while (cur_tag != NULL)
    {
        /* CVE-2007-0071 */
        if (cur_tag->TagCode == TYPE_DefineSceneAndFrameLabelData)
        {
            detect_cursor = cur_tag->tag_data;

            if (cur_tag->Length >= 5)
            {
                if (detect_cursor[0] & 0x80 && detect_cursor[1] & 0x80
                    && detect_cursor[2] & 0x80 && detect_cursor[3] & 0x80
                    && detect_cursor[4] & 0x08)
                {

                    DEBUG_SWF(printf("SceneCount[0] = 0x%02x, [1] = 0x%02x, [2] = 0x%02x,  [3] = 0x%02x,  [4] = 0x%02x\n",
                                     detect_cursor[0], detect_cursor[1], detect_cursor[2], detect_cursor[3], detect_cursor[4]);)

                    *detectID = SWFDetectionID_CVE_2007_0071;
                    return SWF_OK;
                }
            }
            else
                return SWF_ERROR_NODATA;
        }

        /* CVE-2007-5400 */
        /* some compreessed SWF file ends up with extra data after End tag, we need to ignore them */
        if (cur_tag->TagCode == TYPE_ShowFrame && !endtag_found)
            count_showframe++;

        if (cur_tag->TagCode == TYPE_End)
            endtag_found = 1;

        /* CVE-2005-2628 */
        if (cur_tag->TagCode == TYPE_DoAction)
        {
            sub_error = scanActionRecords(cur_tag, detectID);

            if (sub_error != SWF_OK || *detectID != SWFDetectionID_NONE)
                return sub_error;
        }

        if (cur_tag->TagCode == TYPE_DefineSprite && cur_tag->tags.tag_head != NULL)
        {
            if (cur_tag->Length >= 6)   // 6 is minium size of DefineSprite
            {
                tmp_cursor = cur_tag->tag_data + 2;
                definesprite_framecount = READ_LITTLE_16(tmp_cursor);

                sub_error = scanTAGs(&(cur_tag->tags), detectID, definesprite_framecount);

                if (sub_error != SWF_OK || *detectID != SWFDetectionID_NONE)
                    return sub_error;
            }
        }

        cur_tag = cur_tag->next;
    }

    /* CVE-2007-5400 */
    /* CVE-2006-0323 */
    if (frame_count != count_showframe)
    {
        DEBUG_SWF(printf("Header.FrameCount [0x%08x], actual count [0x%08x]\n", frame_count, count_showframe);)
        *detectID = SWFDetectionID_CVE_2007_5400;
        return SWF_OK;
    }

    return SWF_OK;
}

static void initAlert(BLOCK_META_DATA *metaData, ALERT *alert)
{
    alert->event_id   = metaData->eventid;
    alert->priority   = 1;
    if (inet_ntop(AF_INET, &metaData->src_ip, alert->src_ip, sizeof(alert->src_ip)) == NULL)
        alert->src_ip[0] = 0;
    if (inet_ntop(AF_INET, &metaData->dst_ip, alert->dst_ip, sizeof(alert->dst_ip)) == NULL)
        alert->dst_ip[0] = 0;
    alert->ip_proto   = metaData->ip_proto;
    alert->src_port   = metaData->src_port;
    alert->dst_port   = metaData->dst_port;
    alert->msg        = NULL;
    alert->msg_size   = 0;
    alert->short_data = NULL;
    alert->sd_size    = 0;
    uuid_copy(alert->dataType, SWF_FILE);
    if ((alert->main_md5 = malloc(MD5_SIZE)) == NULL)
    {
        printf("Failed to malloc a MD5\n");
        exit(-1);
    }
    md5sum(metaData->data, metaData->size, alert->main_md5);
    alert->long_data  = NULL;
    alert->ld_size    = 0;
    alert->data_block = metaData->data;
    alert->db_size    = metaData->size;
    alert->norm_block = NULL;
    alert->norm_size  = 0;
}

static void clearAlert(ALERT *alert)
{
    free(alert->main_md5);
}

/* any anomalies found during parsing, report them */
static void sendWarnings(ALERT *alert, ErrorCode errCode)
{
    const char *warning_msg;

    switch (errCode)
    {
        case SWF_ERROR_INVALID_SIG:
            warning_msg = "Warning: SWF signature is not valid";
            break;
        case SWF_ERROR_NODATA:
        case SWF_ERROR_INVALID_TAGLENGTH:
            warning_msg = "Warning: The inspected file may be truncated";
            break;
        case SWF_ERROR_MEMORY:
            warning_msg = "Warning: There is not memory related error";
            break;
        case SWF_ERROR_DECOMPRESSION:
            warning_msg = "Warning: Decompression failed";
            break;
        default:
            warning_msg = "Warning: unhandled warning";
    }

#ifdef TEST
    printf("%s\n",  warning_msg);
#else
#if 0
    gDetectionAPI->sendAlert(SWFDetectionID_Warnings, warning_msg, 0, 0, NULL);
#else
    alert->msg = warning_msg;
    alert->msg_size = strlen(warning_msg);
    sendAlert(alert);
#endif
#endif
}

static void sendDetectionResult(ALERT *alert, SWFDetectionID detectID, ErrorCode errCode)
{
    if (detectID == SWFDetectionID_NONE && errCode == SWF_OK)
    {
        printf("Scanned SWF file is clean\n");  // TODO: send result to dispatcher
    }
    else if (detectID < SWFDetectionID_END)
    {
#ifdef TEST
        printf("%s\n", DetectionResults[detectID]->description);
#else
#if 0
        gDetectionAPI->sendAlert(detectID, DetectionResults[detectID]->description, 0, 0, NULL);
#else
        alert->msg = DetectionResults[detectID]->description;
        alert->msg_size = strlen(DetectionResults[detectID]->description);
        sendAlert(alert);
#endif
#endif
    }
}

static void scanSWF(BLOCK_META_DATA *metaData)
{
    SWFFileInfo    swffile;
    ErrorCode      errCode;
    ALERT          alert;
    SWFDetectionID detectID = SWFDetectionID_NONE;

    unsigned char *cursor = metaData->data;
    uint32_t       size   = metaData->size;

    if (cursor == NULL || size == 0 || size > 0x7FFFFFFF)
    {
        printf("Invalid data\n");
        return;
    }

    memset(&swffile, 0, sizeof(SWFFileInfo));

    swffile.bof = cursor;
    if (cursor + size <= cursor)
    {
        DEBUG_SWF(printf("ERROR: file size[0x%08x] is too big\n", size);)
        return;
    }

    initAlert(metaData, &alert);
    swffile.eof = cursor + size;

    errCode = parseSWF(&swffile, cursor, size, &alert);

    if (errCode != SWF_OK)
        sendWarnings(&alert, errCode);

    errCode = scanTAGs(&(swffile.tags), &detectID, swffile.header.FrameCount);

    if (errCode != SWF_OK)
        sendWarnings(&alert, errCode);

    /* if some legitimate tags are scanned */
    if (swffile.tags.tag_head != NULL)
        sendDetectionResult(&alert, detectID, errCode);

    /* epilog */
    freeTAGs(&(swffile.tags));

    if (swffile.compressed)
        free(swffile.decompressed_data);

    clearAlert(&alert);
}

SO_PUBLIC HRESULT initNug(DetectionAPI *detection)
{
    uuid_t list;

    uuid_copy(list, SWF_FILE);

    gDetectionAPI = detection;

#if 0
    if (gDetectionAPI->registerHandler == NULL ||
        gDetectionAPI->sendAlert       == NULL
        /* gDetectionAPI->sendData        == NULL || */
/*         gDetectionAPI->sendMetaData    == NULL */)
    {
        return R_FAIL;
    }
#endif

    if (gDetectionAPI->registerHandler != NULL)
        gDetectionAPI->registerHandler(&scanSWF, (const uuid_t *)&list, 1, TESTDETECT2);
    else
    {
        printf("registerHandler is not valid\n");
        return R_FAIL;
    }

    return R_SUCCESS;
}

